#include "laser.h"
#include "pt_cornell_rp2040_v1.h"
// PWM wrap value and clock divide value
// For a CPU rate of 125 MHz, this gives
// a PWM frequency of 1 kHz.
// 20ms period = 1,250,000 cycles/s * .02s =  2500
#define WRAPVAL 25000
// 50Hz clk = 125MHz/100 = 1.25MHz
#define CLKDIV 100.0f
int distance = 10000;
float first_angle = 0;
float second_angle = 0;
int reversing = 0;
bool first_captured = false;
bool second_captured = false;
// Variable to hold PWM slice number
uint slice_num;
uint slice_num_motor;
int running = 0;
// PWM duty cycle
volatile int control;
volatile int old_control;
void on_pwm_wrap_motor(){
    if(gpio_get(BTN_FORWARD)){
        pwm_set_chan_level(slice_num_motor, PWM_CHAN_A, WRAPVAL);
        pwm_set_chan_level(slice_num_motor, PWM_CHAN_B, 0);
    }
    else if(gpio_get(BTN_BACKWARD)){
        pwm_set_chan_level(slice_num_motor, PWM_CHAN_B, WRAPVAL);
        pwm_set_chan_level(slice_num_motor, PWM_CHAN_A, 0);
    }
    else if(!running){
        pwm_set_chan_level(slice_num_motor, PWM_CHAN_B, 0);
        pwm_set_chan_level(slice_num_motor, PWM_CHAN_A, 0);
    }
    pwm_clear_irq(pwm_gpio_to_slice_num(TRIGGER_FORWARD));

}
void set_angle(float angle)
{
    // 1250 = -90
    // 1875 = 0
    // 2500 = 90

    // clamp angle at either -90 or 90 degrees
    if (angle > 90)
        angle = 90;
    else if (angle < -90)
        angle = -90;

    control = 1875 + 6.94 * 2 * angle;
    pwm_set_chan_level(slice_num, PWM_CHAN_B, control);
}
static PT_THREAD(protothread_sweep(struct pt *pt))
{
    #define DIST_MAX 4000
    #define VISUAL_MAX 1500
    PT_BEGIN(pt);
    static int start = -90;
    static int end = 90;
    static float delay = 45;
    static int i = -45;
    static int dist_max_counter = 0;
    static int reverse = false;
    static int Num;
    set_angle(start);
    if (delay > 0)
    {
        while(1)
        {
            if (i >= end) reverse = true;
            else if (i <= start) reverse = false;

            if (first_captured && second_captured)
            {
                // Set to average angle
                Num = (first_angle + second_angle)/2;
                
                set_angle(Num);
                i = Num;
      
                running = 1;
                pwm_set_chan_level(slice_num_motor, PWM_CHAN_A, WRAPVAL);
                pwm_set_chan_level(slice_num_motor, PWM_CHAN_B, 0);
                PT_YIELD_usec(200000);
                running = 0;
                
                // If the object moves, then the servo will start sweeping again
                if (distance > DIST_MAX)
                {
                    dist_max_counter ++;
                    if (dist_max_counter >= 8){
                        dist_max_counter = 0;
                        i = (first_angle + second_angle) / 2;
                        first_captured = false;
                        second_captured = false;
                    }
                }
            }
            else
                set_angle(i);

            // Captures first edge of detected object
            if (distance < VISUAL_MAX && !first_captured)
            {
                // Currently commented out while not using ultrasonic sensor
                 first_angle = i;
                 first_captured = true;
            }
            // Captures second edge of detected object
            else if (distance > VISUAL_MAX && first_captured && !second_captured)
            {
                second_angle = i;
                second_captured = true;
            }
            if (reverse) i--;
            else i++;

            PT_YIELD_usec(delay * 1000);
        }
    }
    PT_END(pt);
}


static PT_THREAD(protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
    sprintf(pt_serial_out_buffer, "distance: %d \n", distance);
    serial_write;
    if(DEBUG){
        sprintf(pt_serial_out_buffer, "\n%s\n", buf);
        serial_write;
    }

    sprintf(pt_serial_out_buffer, "%d", first_captured);
    serial_write;
    sprintf(pt_serial_out_buffer, "   %d", second_captured);
    serial_write;
    sprintf(pt_serial_out_buffer, "   %d", running);
    serial_write;
    sprintf(pt_serial_out_buffer, "\n\r");
    serial_write;
    PT_YIELD_usec(100000);
    PT_END(pt);
}

static PT_THREAD(protothread_laser(struct pt *pt))
{
    PT_BEGIN(pt);
    laser_read(&distance);
    PT_YIELD_usec(100000);
    PT_END(pt);
}
int main()
{

    // Initialize stdio
    stdio_init_all();
    laser_init();
    gpio_init(MOTION_SENSE_1);
    gpio_set_dir(MOTION_SENSE_1, GPIO_IN);
    gpio_init(MOTION_SENSE_2);
    gpio_set_dir(MOTION_SENSE_2, GPIO_IN);
    gpio_init(MOTION_SENSE_3);
    gpio_set_dir(MOTION_SENSE_3, GPIO_IN);
    gpio_init(BTN_FORWARD);
    gpio_set_dir(BTN_FORWARD, GPIO_IN);
    gpio_init(BTN_BACKWARD);
    gpio_set_dir(BTN_BACKWARD, GPIO_IN);

    ////////////////////////////////////////////////////////////////////////
    ///////////////////////// PWM CONFIGURATION ////////////////////////////
    ////////////////////////////////////////////////////////////////////////
    // Tell GPIO 5 that it is allocated to the PWM
    gpio_set_function(SERVO, GPIO_FUNC_PWM);
    gpio_set_function(TRIGGER_FORWARD, GPIO_FUNC_PWM);
    gpio_set_function(TRIGGER_BACKWARD, GPIO_FUNC_PWM);
    // Find out which PWM slice is connected to GPIO 5 (it's slice 2)
    slice_num = pwm_gpio_to_slice_num(SERVO);
    slice_num_motor = pwm_gpio_to_slice_num(TRIGGER_FORWARD);

    // Mask our slice's IRQ output into the PWM block's single interrupt line,
    // and register our interrupt handler
    //pwm_clear_irq(slice_num);
    //pwm_set_irq_enabled(slice_num, true);
    //irq_set_exclusive_handler(PWM_IRQ_WRAP, on_pwm_wrap);
    //irq_set_enabled(PWM_IRQ_WRAP, true);

    pwm_clear_irq(slice_num_motor);
    pwm_set_irq_enabled(slice_num_motor, true);
    irq_set_exclusive_handler(PWM_IRQ_WRAP, on_pwm_wrap_motor);
    irq_set_enabled(PWM_IRQ_WRAP, true);

    // This section configures the period of the PWM signals
    pwm_set_wrap(slice_num, WRAPVAL);
    pwm_set_clkdiv(slice_num, CLKDIV);
    pwm_set_wrap(slice_num_motor, WRAPVAL);
    pwm_set_clkdiv(slice_num_motor, CLKDIV);

    // This sets duty cycle
    pwm_set_chan_level(slice_num, PWM_CHAN_B, 250);

    // Start the channel
    pwm_set_enabled(slice_num,1);
    pwm_set_enabled(slice_num_motor,1);

    ////////////////////////////////////////////////////////////////////////
    ///////////////////////////// ROCK AND ROLL ////////////////////////////
    ////////////////////////////////////////////////////////////////////////
    pt_add_thread(protothread_serial);
    pt_add_thread(protothread_laser);
    pt_add_thread(protothread_sweep);
    pt_schedule_start;
}
